from flask_autoapi.form.field import Field


class ParamMetaclass(type):
    def __new__(cls, name, bases, attrs):
        fields = dict()
        for k, field in attrs.items():
            if isinstance(field, Field):
                fields[k] = field
                field.name = k
        attrs["fields"] = fields
        return type.__new__(cls, name, bases, attrs)


class Param(metaclass=ParamMetaclass):
    def __init__(self, **kwargs):
        fields = getattr(self, "fields")
        for field in fields.values():
            setattr(self, field.name, kwargs.get(field.name))
    
    def validate(self) -> bool:
        return True

    def json(self):
        return self.__dict__



import math
from flask import request
from flask_restful import Resource, marshal_with

from flask_autoapi.utils.response import JsonResponse, resource_fields
from flask_autoapi.utils.message import BAD_REQUEST, OBJECT_SAVE_FAILED

class BaseEndpoint(Resource):

    Model = None
    Type = None
    # 由于 method_decorators 是内层装饰器，比 decorators 先起作用，所以一律使用 method_decorators
    decorators = []
    method_decorators = {
        "get":      [marshal_with(resource_fields)], 
        "put":      [marshal_with(resource_fields)],
        "post":     [marshal_with(resource_fields)],
        "delete":   [marshal_with(resource_fields)],
        "options":  [marshal_with(resource_fields)],
    }

    @classmethod
    def add_decorators(cls, decorator_list):
        if not decorator_list:
            return
        if not isinstance(decorator_list, (list, tuple)):
            raise Exception("TypeError, decorator_list should be list or tuple, but {} found".format(type(decorator_list)))
        cls.decorators += decorator_list
    
    @classmethod
    def add_method_decorators(cls, method_decorators):
        if not method_decorators:
            return
        if not isinstance(method_decorators, dict):
            raise Exception("TypeError, method_decorators should be dict, but {} found, endpoint = {}".format(type(method_decorators), cls))
        for key, value in method_decorators.items():
            cls.method_decorators[key] += value

    def get(self, id):
        """
        @api {GET} /{{project_name}}/{{ModelName.lower()}}/:id 获取{{Title}}详情
        @apiName Get{{ModelName}}
        @apiGroup {{Group}}

        @apiExample 返回值
{{ DATA}}

        """
        without_fields = request.args.get("without_fields")
        without_fields = without_fields.split(",") if without_fields else None
        r = self.Model.get_with_pk(id, without_fields)
        if not r:
            return JsonResponse()
        r.get_method_fields()
        r = self.Model.to_json(r, without_fields) if r else None
        r = self.Model.out_handlers(**r)
        r = self.Model.diy_after_get(**r)
        return JsonResponse(data=r)
    
    def post(self):
        """
        @api {POST} /{{project_name}}/{{ModelName.lower()}} 创建{{Title}}
        @apiName Create{{ModelName}}
        @apiGroup {{Group}}

        @apiExample 参数
        {%- for field in Fields %}
        {{ str_align(standard_type(field.field_type))}} \t {{ str_align(field.name, 15) }}  # {% if field.null is sameas true %} 非必填项 {% else %} 必填项 {% endif %} {% if field.verbose_name %}, {{field.verbose_name}} {% endif %}{% endfor %}
        

        @apiExample 返回值
{{ DATA}}

        """
        params = request.get_json() if request.content_type == "application/json" else request.form.to_dict()
        params.update(request.files.to_dict())
        params = self.Model.in_handlers(**params)
        params = self.Model.format_params(**params)
        status = self.Model.verify_params(**params)
        if not status:
            return JsonResponse(BAD_REQUEST)
        params = self.Model.upload_files(**params)
        if not self.Model.validate(**params):
            return JsonResponse(BAD_REQUEST)
        self.Model.diy_before_save(**params)
        r = self.Model.create(**params)
        self.Model.diy_after_save(r)
        r.mtom(**params)
        r.get_method_fields()
        r = self.Model.to_json(r) if r else None
        r = self.Model.out_handlers(**r)
        r = self.Model.diy_after_get(**r)
        return JsonResponse(data=r)
    
    def put(self, id):
        """
        @api {PUT} /{{project_name}}/{{ModelName.lower()}}/:id 更新{{Title}}
        @apiName Update{{ModelName}}
        @apiGroup {{Group}}

        @apiExample 参数
        {%- for field in Fields %}
        {{ str_align(standard_type(field.field_type))}} \t {{ str_align(field.name, 15) }}  # {% if field.null is sameas true %} 非必填项 {% else %} 必填项 {% endif %} {% if field.verbose_name %}, {{field.verbose_name}} {% endif %}{% endfor %}

        @apiExample 返回值
{{ DATA}}
        
        """
        params = request.get_json() if request.content_type == "application/json" else request.form.to_dict()
        params.update(request.files.to_dict())
        params = self.Model.in_handlers(**params)
        params = self.Model.format_params(**params)
        status = self.Model.verify_params(**params)
        if not status:
            return JsonResponse(BAD_REQUEST)
        params = self.Model.upload_files(**params)
        if not self.Model.validate(**params):
            return JsonResponse(BAD_REQUEST)
        self.Model.diy_before_save(**params)
        r = self.Model.update_by_pk(id, **params)
        self.Model.diy_after_save(r)
        r.mtom(**params)
        r.get_method_fields()
        r = self.Model.to_json(r) if r else None
        r = self.Model.out_handlers(**r)
        r = self.Model.diy_after_get(**r)
        return JsonResponse(data=r)
    
    def delete(self, id):
        """
        @api {DELETE} /{{project_name}}/{{ModelName.lower()}}/:id 删除{{Title}}
        @apiName Delete{{ModelName}}
        @apiGroup {{Group}}

        @apiExample 返回值
        {
            "code": 0,
            "message": "",
            "data": None,
        }
        
        """
        self.Model.delete().where(self.Model._meta.primary_key == id).execute()
        return JsonResponse()


class BaseListEndpoint(Resource):

    Model = None
    Type = "List"
    decorators = []
    method_decorators = {
        "get":      [marshal_with(resource_fields)], 
        "put":      [marshal_with(resource_fields)],
        "post":     [marshal_with(resource_fields)],
        "delete":   [marshal_with(resource_fields)],
        "options":  [marshal_with(resource_fields)],
    }

    @classmethod
    def add_decorators(cls, decorator_list):
        if not decorator_list:
            return
        if not isinstance(decorator_list, (list, tuple)):
            raise Exception("格式错误")
        cls.decorators += decorator_list
    
    @classmethod
    def add_method_decorators(cls, method_decorators):
        if not method_decorators:
            return
        if not isinstance(method_decorators, dict):
            raise Exception("TypeError, method_decorators should be dict, but {} found".format(type(method_decorators)))
        for key, value in method_decorators.items():
            cls.method_decorators[key] += value

    def get(self):
        """
        @api {GET} /{{project_name}}/{{ModelName.lower()}}/list?page=2&page_size=10&order=0 获取{{Title}}列表
        @apiName Get{{ModelName}}List
        @apiGroup {{Group}}

        @apiExample 参数
        int    page    # 页码。非必填，默认1。
        int    page_size     # 每页数量。非必填，默认10。
        int    order   # 排序方法。非必填，默认0。0表示按时间倒序，1表示按时间顺序。

        @apiExample 返回值
{{LIST_DATA}}
        """
        args = request.args.to_dict()
        try:
            page  = int(args.get("page", 1))
            page_size   = int(args.get("page_size", 10))
            order = int(args.get("order", 0))
        except:
            return JsonResponse(BAD_REQUEST)
        if not order in (0, 1):
            return JsonResponse(BAD_REQUEST)
        args = self.Model.verify_list_args(**args)
        without_fields = request.args.get("without_fields")
        without_fields = without_fields.split(",") if without_fields else None
        fields = self.Model.get_fields()
        fields = [field for field in fields if field.name not in without_fields] \
                    if without_fields else fields
        result = self.Model.select(*fields)
        for key, value in args.items():
            if value.find("~") >= 0:
                a, b = value.split("~")
                a = a.strip()
                b = b.strip()
                if a:
                    result = result.where(getattr(self.Model, key) >= a)
                if b:
                    result = result.where(getattr(self.Model, key) <= b)
            else:
                result = result.where(getattr(self.Model, key) == value)
        result = result.order_by(self.Model.create_time.desc()) if order == 0 else result.order_by(self.Model.create_time.asc())
        total_count = result.count()
        result = result.offset((page-1)*page_size).limit(page_size)
        result = [self.Model.to_json(r.get_method_fields(), without_fields) for r in result] if result else None
        result = [self.Model.out_handlers(**r) for r in result] if result else None
        result = [self.Model.diy_after_get(**r) for r in result] if result else None
        result = {
            "page_size":page_size,
            "total": total_count,
            "page": page,
            "result":result
        }
        return JsonResponse(data=result)
        


import uuid
import hashlib
from decimal import Decimal
from datetime import datetime

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"


def datetime_to_str(_datetime, _format=DATETIME_FORMAT):
    return _datetime.strftime(_format)

def field_to_json(value, datetime_format=DATETIME_FORMAT):
    ret = value
    if isinstance(value, datetime):
        ret = datetime_to_str(value, datetime_format)
    elif isinstance(value, list):
        ret = [field_to_json(_) for _ in value] if value else None
    elif isinstance(value, dict):
        ret = {k: field_to_json(v) for k, v in value.items()} if value else None
    elif isinstance(value, bytes):
        ret = value.decode("utf-8")
    elif isinstance(value, bool):
        ret = int(ret)
    elif isinstance(value, uuid.UUID):
        ret = str(value)
    elif isinstance(value, Decimal):
        ret = float(ret)
    return ret

def content_md5(content):
    hash_md5 = hashlib.md5(content)
    return hash_md5.hexdigest()



import os

from flask_autoapi.utils.diyutils import content_md5
from flask_autoapi.storage.handler import (FileHandler, MinioHandler,
                                           QiniuHandler, TencentHandler)
from flask_autoapi.storage.client import Client, MinioClient, QiniuClient, TencentClient


class Storage(object):
    def __init__(self, client: Client, bucket):
        if not isinstance(client, Client):
            raise TypeError("client should be type of Client")
        self.client = client
        self.bucket = bucket
        self.handler = self.client.handler_class(
            client=self.client,
            bucket=self.bucket,
        )

    def write(self, file_obj):
        if not file_obj:
            return
        self._file_attrs(file_obj)
        return self.handler.write(file_obj)

    def read(self, md5_hash):
        if not md5_hash:
            return
        return self.handler.read(md5_hash)

    def _file_attrs(self, file_obj):
        if not hasattr(file_obj, "md5_hash"):
            content = file_obj.read()
            setattr(file_obj, "md5_hash", content_md5(content))
            setattr(file_obj, "length", len(content))
            file_obj.seek(0, os.SEEK_SET)





class Field(object):
    field_type = None

    def __init__(self, verbose_name="", default=None, null=False):
        self.null = null
        self.name = ""
        self.default = default
        self.verbose_name = verbose_name


class StringField(Field):
    field_type = "string"


class IntegerField(Field):
    field_type = "int"


class FloatField(Field):
    field_type = "float"


class BooleanField(Field):
    field_type = "bool"


class ListField(Field):
    field_type = "list"


class DictField(Field):
    field_type = "dict"


import os
import requests

from flask_autoapi.utils.diyutils import content_md5


class StorageHandler(object):
    def __init__(self, client, bucket):
        self.client = client
        self.bucket = bucket


class FileHandler(StorageHandler):
    def write(self, file_obj):
        content = str(file_obj.read(), encoding="latin-1")
        md5_hash = content_md5(content.encode("latin-1"))
        file_path = os.path.join(self.bucket, md5_hash)
        with open(file_path, "wb") as f:
            f.write(bytes(content, encoding="latin-1"))
        return md5_hash

    def read(self, md5_hash):
        file_path = os.path.join(self.bucket, md5_hash)
        with open(file_path, "rb") as f:
            content = f.read()
        return content


class MinioHandler(StorageHandler):
    def write(self, file_obj):
        etag = None
        try:
            etag = self.client.put_object(
                bucket_name=self.bucket,
                object_name=file_obj.md5_hash,
                data=file_obj,
                length=file_obj.length,
            )
        except Exception as e:
            print("上传文件到 minio 失败，{}".format(e))
        return etag

    def read(self, md5_hash):
        r = self.client.get_object(
            bucket_name=self.bucket,
            object_name=md5_hash,
        )
        return r.read()


class QiniuHandler(StorageHandler):
    def write(self, file_obj):
        content = file_obj.read()
        token = self.client.upload_token(
            bucket=self.bucket,
            key=file_obj.md5_hash,
            expires=10,
        )
        r, info = self.client.put_data(
            token=token,
            key=file_obj.md5_hash,
            data=content,
        )
        return r["key"]

    def read(self, md5_hash):
        base_url = "http://{BUCKET_URL}/{KEY}".format(
            BUCKET_URL=self.bucket,
            KEY=md5_hash,
        )
        private_url = self.client.private_download_url(
            base_url,
            expires=10,
        )
        r = requests.get(private_url)
        if r.status_code != 200:
            raise Exception("获取七牛文件失败, status_code = {}".format(r.status_code))
        return r.content


class TencentHandler(StorageHandler):
    """腾讯云对象存储"""
    def write(self, file_obj):
        self.client.put_object(
            Bucket=self.bucket,
            Body=file_obj,
            Key=file_obj.md5_hash,
        )

    def read(self, md5_hash):
        r = self.client.get_object(
            Bucket=self.bucket,
            Key=md5_hash,
        )
        return r["Body"].get_stream_to_file(md5_hash)


from flask_restful import fields
from flask_autoapi.utils.message import MESSAGE


resource_fields = {
    "code" : fields.Integer,
    "message" : fields.String,
    "data" : fields.Raw(default=None)
}

class JsonResponse(object):

    msg = MESSAGE

    def __init__(self, code=0, message="", data=None):
        if not message:
            message = self.msg.get(code)
        self.code = code
        self.data = data
        self.message = message
    
    @classmethod
    def set_msg(cls, msg):
        cls.msg = msg




import uuid
import types
import werkzeug
from io import BytesIO
from playhouse.shortcuts import model_to_dict
from peewee import (Model, CharField, ManyToManyField, ForeignKeyField, Field,
                    FieldAccessor, MetaField, Metadata, AutoField,
                    DateTimeField, IntegerField, MySQLDatabase)

from flask_autoapi.storage import Storage
from flask_autoapi.utils.diyutils import field_to_json, content_md5

MySQLDatabase.field_types["OBJECT"] = "VARCHAR(40)"


def api_model_to_dict(obj, **kwargs):
    data = model_to_dict(obj, **kwargs)
    method_fields = list(obj._meta.method_fields.values())
    for field in method_fields:
        data[field.name] = getattr(obj, field.name)
    return data


class MethodClass(object):
    def get(self, obj):
        raise NotImplementedError("get function should be implemented.")

    def get_example(self):
        raise NotImplementedError(
            "get_example function should be implemented.")


class ApiMethodField(MetaField):
    """
    A readonly field, get value by object method.
    """
    field_type = "METHOD"

    def __init__(self, method_class, verbose_name=None):
        self.choices = None
        self.read_only = True
        self.method_class = method_class
        self.verbose_name = verbose_name
        if not self.method_class:
            raise Exception("method_class should not be None.")
        if not issubclass(self.method_class, MethodClass):
            raise Exception(
                "method_class should not be subclass of MethodClass.")


class ApiMetadata(Metadata):
    def __init__(self, model, **kwargs):
        self.method_fields = {}
        super(ApiMetadata, self).__init__(model, **kwargs)

    def add_field(self, field_name, field, set_attribute=True):
        super(ApiMetadata, self).add_field(field_name, field, set_attribute)
        if isinstance(field, ApiMethodField) and field.name:
            self.add_method_field(field)

    def add_method_field(self, field):
        self.method_fields[field.name] = field


class ApiModel(Model):
    @classmethod
    def set_meta(cls, **kwargs):
        for key, value in kwargs.items():
            setattr(cls._meta, key, value)

    @classmethod
    def get_with_pk(cls, pk_value, without_field_names=None):
        if without_field_names and not isinstance(without_field_names,
                                                  (list, tuple)):
            return None
        fields = cls.get_fields()
        fields = [field for field in fields if field.name not in without_field_names] \
                    if without_field_names else fields
        return cls.select(*fields).where(
            cls._meta.primary_key == pk_value).first()

    @classmethod
    def get_field_names(cls):
        return cls._meta.sorted_field_names

    @classmethod
    def get_all_field_names(cls):
        all_fields = cls.get_fields() + list(cls._meta.manytomany.values())
        return [field.name for field in all_fields]

    @classmethod
    def get_fields(cls):
        fields = []
        for field_name in cls.get_field_names():
            field = cls._meta.fields[field_name]
            fields.append(field)
        return fields

    @classmethod
    def get_display_fields(cls):
        fields = cls.get_fields()
        fields = [field for field in fields if not field._hidden]
        return fields

    @classmethod
    def get_post_fields(cls):
        fields = cls.get_fields()
        fields = [
            field for field in fields if not getattr(field, "read_only", False)
        ]
        return fields

    @classmethod
    def get_field_by_name(cls, field_name):
        if field_name in cls._meta.fields:
            return cls._meta.fields[field_name]
        return None

    @classmethod
    def in_handlers(cls, **params):
        # 对原始的参数进行处理
        fields = cls.get_fields() + list(cls._meta.manytomany.values())
        for field in fields:
            if not (hasattr(field, "in_handler") and field.in_handler):
                continue
            handler = getattr(cls, field.in_handler)
            if not (handler and isinstance(handler, types.MethodType)):
                raise Exception(
                    "in_handler should be function, but {} found.".format(
                        type(handler)))
            params[field.name] = handler(params.get(field.name))
            if field.name in cls._meta.manytomany:
                params[field.name + "_value"] = params.pop(field.name)
        return params

    def mtom(self, **params):
        for field_name, field in self._meta.manytomany.items():
            ins = getattr(self, field_name)
            ins.clear()
            value = params[field_name + "_value"]
            if isinstance(value, (list, tuple)):
                for v in value:
                    ins.add(v)
            else:
                ins.add(value)

    @classmethod
    def out_handlers(cls, **data):
        # 对 to_json() 之后的结果进行处理
        fields = cls.get_fields() + list(cls._meta.manytomany.values())
        for field in fields:
            if not (hasattr(field, "out_handler")
                    and getattr(field, "out_handler")):
                continue
            handler = getattr(cls, field.out_handler)
            if not (handler and isinstance(handler, types.MethodType)):
                raise Exception(
                    "out_handler should be MethodType, but {} found.".format(
                        type(handler)))
            data[field.name] = handler(data.get(field.name))
        return data

    @classmethod
    def format_params(cls, **params):
        # all_field_names = cls.get_all_field_names()
        # for key in list(params.keys()):
        #     if key in all_field_names:
        #         continue
        #     params.pop(key)
        # 格式化参数，主要是将 str 转换成 file 对象
        fields = cls.get_fields()
        for field in fields:
            if isinstance(field, ApiFileIDField):
                if field.source_type == "string":
                    content = params.get(field.source_name)
                    if not content:
                        continue
                    if not isinstance(content, (str, bytes)):
                        raise Exception("{} 应该为 str 或 bytes，而不是 {}".format(
                            field.name, type(content)))
                    if isinstance(content, str):
                        content = content.encode("utf8")
                    f = BytesIO(content)
                    setattr(f, "md5_hash", content_md5(content))
                    setattr(f, "length", len(content))
                    params[field.name] = f
                elif field.source_type == "file":
                    f = params.get(field.source_name)
                    if not isinstance(
                            f,
                        (werkzeug.datastructures.FileStorage, type(None))):
                        raise Exception(
                            "类型错误，{} 应该为 werkzeug.datastructures.FileStorage 类型，而不是 {}"
                            .format(field.source_name, type(f)))
                    params[field.name] = f
                else:
                    raise Exception("不能识别的类型, 无法转换，source_type = {}".format(
                        field.source_type))
        return params

    def get_method_fields(self):
        fields = list(self._meta.method_fields.values())
        for field in fields:
            r = field.method_class.get(self)
            setattr(self, field.name, r)
        return self

    @classmethod
    def verify_params(cls, **params):
        # 验证 params 中的参数，主要验证非 None 字段是否有值
        fields = cls.get_fields()
        for field in fields:
            if field.auto_increment or field.default is not None or field.null:
                continue
            if params.get(field.name) is None:
                print("{} is None".format(field.name))
                return False
        return True

    @classmethod
    def upload_files(cls, **params):
        # cls.init_storage()
        fields = cls.get_fields()
        for field in fields:
            if not params.get(field.name):
                continue
            if isinstance(field, ApiFileIDField):
                file_id = cls._meta.storage.write(params[field.name])
                params[field.name] = file_id
        return params

    @classmethod
    def verify_list_args(cls, **args):
        # 验证 list 接口的参数
        result = {}
        for key, value in args.items():
            if not key in cls._meta.filter_fields:
                continue
            result[key] = value
        return result

    @classmethod
    def get_fileid_field_name(cls, **params):
        fields = cls.get_fields()
        for field in fields:
            if isinstance(field,
                          ApiFileIDField) and not params.get(field.name):
                return field.name
        return None

    @classmethod
    def update_by_pk(cls, pk_value, **params):
        status = cls.verify_params(**params)
        if not status:
            return
        field_names = cls.get_field_names()
        r = cls.get_with_pk(pk_value)
        for key, value in params.items():
            if not key in field_names or cls.get_field_by_name(
                    key).primary_key:
                continue
            setattr(r, key, value)
        r.save()
        return r

    @classmethod
    def validate(cls, **params):
        """
        validate 用于验证参数
        返回 True 表示正确；
        返回 False 表示错误，会返回 BadRequest
        """
        return True

    @classmethod
    def diy_before_save(cls, **params):
        """
        用于 POST/PUT 方法。
        做一些自定义操作，但是不能修改参数。
        修改参数应该使用各个Field 的 in_handler。                        
        """
        return True

    @classmethod
    def diy_after_save(cls, obj):
        """
        用于 POST/PUT 方法。
        create/save/update 之后，自定义操作。
        """
        return True

    @classmethod
    def diy_after_get(cls, **json_data):
        """
        在 GET 方法中获取到对象并转换为 json 之后，对返回值做一些操作。
        """
        return json_data

    @classmethod
    def to_json(cls,
                obj,
                without_fields=None,
                datetime_format="%Y-%m-%d %H:%M:%S"):
        fields = cls.get_fields()
        # for field in fields:
        #     # 如果数据源为 string，则返回时也应该返回 string
        #     if hasattr(field, "source_type") and field.source_type == "string" and getattr(obj, field.name):
        #         content = cls._meta.storage.read(getattr(obj, field.name))
        #         setattr(obj, field.name, content)
        # to json
        r = api_model_to_dict(obj, manytomany=cls._meta.manytomany)
        # 将 many-to-many 的数据取出来
        # for field_name, field in cls._meta.manytomany.items():
        #     r[field_name] = [model_to_dict(x) for x in getattr(obj, field_name)]
        # 过滤掉不需要的字段
        result = {}
        for name, value in r.items():
            field = cls.get_field_by_name(name)
            if field._hidden:
                continue
            if without_fields and name in without_fields:
                continue
            result[name] = field_to_json(value, datetime_format)
        return result

    class Meta:
        model_metadata_class = ApiMetadata
        group = ""
        # verbose_name 指定别名，用于显示在 API 文档上。默认为 Model 的名称
        verbose_name = ""
        # filter_fields 用于指定 list 接口的参数
        filter_fields = ()
        api_methods = ("GET", "POST", "PUT", "DELETE", "LIST")
        # api_methods = ()
        api_decorator_list = ()
        api_method_decorators = None
        list_api_decorator_list = ()
        list_api_method_decorators = None
        # storage 是 Storage 的对象
        storage = None
        # store_kind 指定文件存储的方式，支持 file/minio/qiniu
        # store_kind = "file"
        # bucket 指定文件存储文件夹，或云存储的 bucket
        # bucket = ""
        # minio 配置
        # minio_url = ""
        # minio_secure = False
        # minio_access_key = ""
        # minio_secret_key = ""
        # qiniu 配置
        # qiniu_url = ""
        # qiniu_access_key = ""
        # qiniu_secret_key = ""
        # qiniu_bucket_url = ""


class ApiFileIDField(CharField):
    field_type = "OBJECT"

    def __init__(self, max_length=255, *args, **kwargs):
        self.source_name = kwargs.pop("source_name", "file")
        self.source_type = kwargs.pop("source_type", "file")
        super(ApiFileIDField, self).__init__(max_length=max_length,
                                             *args,
                                             **kwargs)

    def db_value(self, file_obj):
        # 写数据库时会使用该函数返回的值，但是并不会修改参数的值
        r = self.model._meta.storage.write(file_obj)
        return r

    def python_value(self, value):
        # just for select sql
        content = self.model._meta.storage.read(value)
        return content


class ApiCharField(CharField):
    def __init__(self, max_length=255, read_only=False, *args, **kwargs):
        self.read_only = read_only
        self.in_handler = kwargs.pop("in_handler", None)
        self.out_handler = kwargs.pop("out_handler", None)
        # if self.in_handler and not isinstance(self.in_handler, types.FunctionType):
        #     raise Exception("in_handler should be function, but {} found.".format(type(self.in_handler)))
        # if self.out_handler and not isinstance(self.out_handler, types.FunctionType):
        #     raise Exception("out_handler should be function, but {} found.".format(type(self.out_handler)))
        super(ApiCharField, self).__init__(max_length=max_length,
                                           *args,
                                           **kwargs)


class ApiManyToManyField(ManyToManyField):
    def __init__(self, model, **kwargs):
        self.in_handler = kwargs.pop("in_handler", None)
        self.out_handler = kwargs.pop("out_handler", None)
        # if self.in_handler and not isinstance(self.in_handler, types.FunctionType):
        #     raise Exception("in_handler should be function, but {} found.".format(type(self.in_handler)))
        # if self.out_handler and not isinstance(self.out_handler, types.FunctionType):
        #     raise Exception("out_handler should be function, but {} found.".format(type(self.out_handler)))
        super(ApiManyToManyField, self).__init__(model, **kwargs)

    def get_through_model(self):
        if not self.through_model:
            lhs, rhs = self.get_models()
            tables = [model._meta.table_name for model in (lhs, rhs)]

            class Meta:
                database = self.model._meta.database
                schema = self.model._meta.schema
                table_name = '%s_%s_through' % tuple(tables)
                indexes = (((lhs._meta.name, rhs._meta.name), True), )

            attrs = {
                lhs._meta.name:
                ForeignKeyField(lhs,
                                on_delete=self._on_delete,
                                on_update=self._on_update),
                rhs._meta.name:
                ForeignKeyField(rhs,
                                on_delete=self._on_delete,
                                on_update=self._on_update)
            }
            attrs['Meta'] = Meta

            self.through_model = type(
                '%s%sThrough' % (lhs.__name__, rhs.__name__), (ApiModel, ),
                attrs)

        return self.through_model


class ApiUUIDField(Field):
    field_type = "char(32)"

    def __init__(self, read_only=False, **kwargs):
        self.read_only = read_only
        self.in_handler = kwargs.pop("in_handler", None)
        self.out_handler = kwargs.pop("out_handler", None)
        # self.obj_func = kwargs.pop("obj_func", None)
        super(ApiUUIDField, self).__init__(**kwargs)

    def db_value(self, value):
        if not isinstance(value, (str, uuid.UUID)):
            return value
        if isinstance(value, str):
            # value = uuid.UUID(value)
            try:
                value = uuid.UUID(value)  # convert hex string to UUID
            except Exception as e:
                print("Failed to convert value {} to uuid".format(value))
        value = value.hex if isinstance(value, uuid.UUID) else str(
            value)  # convert UUID to hex string.
        return value

    def python_value(self, value):
        try:
            value = uuid.UUID(value)  # convert hex string to UUID
        except Exception as e:
            print("Failed to convert value {} to uuid".format(value))
        return value


class ApiAutoField(AutoField):
    def __init__(self, read_only=False, *args, **kwargs):
        self.read_only = read_only
        super(ApiAutoField, self).__init__(*args, **kwargs)


class ApiDateTimeField(DateTimeField):
    def __init__(self, read_only=False, *args, **kwargs):
        self.read_only = read_only
        super(ApiDateTimeField, self).__init__(*args, **kwargs)


class ApiIntegerField(IntegerField):
    def __init__(self, read_only=False, *args, **kwargs):
        self.read_only = read_only
        super(ApiIntegerField, self).__init__(*args, **kwargs)



# 获取所有的 docstring，生成 doc
import os
from flask import request
from functools import wraps
from copy import copy, deepcopy
from flask_restful import Api

from flask_autoapi.model import ApiModel
from flask_autoapi.form.param import Param
from flask_autoapi.endpoint import BaseEndpoint, BaseListEndpoint


class AutoAPI(object):
    def __init__(self):
        self.api = Api()
        self.app = None
        self.model_list = None
        # doc_folder 位于 static 目录下
        self.doc_folder = "docs"
        self._lazy_resources = []
        self._custom_resources = {}
    
    def init_app(self, app, model_list, decorator_list=None, project_name=""):
        if not isinstance(model_list, (list, tuple)):
            raise Exception("model_list 应该是一个列表，不是{}".format(type(model_list)))
        for model in model_list:
            if not issubclass(model, ApiModel):
                raise Exception("model 应该继承自 ApiModel")
        self.app = app
        self.model_list = model_list
        self.project_name = project_name if project_name else os.getcwd().split("/")[-1]
        if not self.project_name:
            raise Exception("project_name 不能为空，需要使用 project_name 作为 URL 前缀")
        self.app.add_url_rule("/docs/", "docs", self._static_file, strict_slashes=False)
        self.app.add_url_rule("/docs/<path:path>", "docs", self._static_file, strict_slashes=False)
        self._add_decorators(decorator_list)
        self._auto_urls()
        self.api.init_app(self.app)
        for data in self._lazy_resources:
            self.update_resource(data[0], *data[1:])
    
    def update_resource(self, resource, *urls):
        self._custom_resources[resource.__name__] = urls[0]
        if self.app:
            endpoint = resource.__name__.lower()
            func = self.api.output(resource.as_view(endpoint))
            self._del_exists_endpoint(endpoint)
            for url in urls:
                self.app.add_url_rule(url, endpoint, func, strict_slashes=False)
        else:
            self._lazy_resources.append((resource, *urls))
    
    def _del_exists_endpoint(self, endpoint):
        if self.app.view_functions.get(endpoint):
            del self.app.view_functions[endpoint]
            self.app.url_map._rules = [rule for rule in self.app.url_map._rules if rule.endpoint != endpoint]
    
    def _static_file(self, path=None):
        if not path:
            path = "index.html"
        path = os.path.join(self.doc_folder, path)
        return self.app.send_static_file(path)
    
    def _add_decorators(self, decorator_list):
        if decorator_list:
            BaseEndpoint.add_decorators(decorator_list)
            BaseListEndpoint.add_decorators(decorator_list)

    def _auto_urls(self):
        endpoints = []
        for model in self.model_list:
            # model._meta.api_methods = [method.upper() for method in model._meta.api_methods]
            class_name = model.__name__ + "Endpoint"
            endpoint = type(class_name, (BaseEndpoint, ), {})
            endpoint.Model = model
            # 无法删除父类的方法，暂时不删除，不给文档即可
            # for method in ["GET", "POST", "PUT", "DELETE"]:
            #     if method not in model._meta.api_methods:
            #         delattr(endpoint, method.lower())
            #         print("del method ", method)
            endpoint.decorators = copy(BaseEndpoint.decorators)
            endpoint.method_decorators = deepcopy(BaseEndpoint.method_decorators)
            endpoint.add_decorators(model._meta.api_decorator_list)
            endpoint.add_method_decorators(model._meta.api_method_decorators)
            endpoints.append(endpoint)

            class_name = model.__name__ + "ListEndpoint"
            endpoint = type(class_name, (BaseListEndpoint, ), {})
            endpoint.Model = model
            endpoint.decorators = copy(BaseListEndpoint.decorators)
            endpoint.method_decorators = deepcopy(BaseListEndpoint.method_decorators)
            endpoint.add_decorators(model._meta.list_api_decorator_list)
            endpoint.add_method_decorators(model._meta.list_api_method_decorators)
            endpoints.append(endpoint)
        
        for endpoint in endpoints:
            if endpoint.Type:
                url = "/".join(["", self.project_name.lower(), endpoint.Model.__name__.lower(), endpoint.Type.lower(), ""])
                self.api.add_resource(endpoint, url, endpoint=endpoint.__name__.lower(), strict_slashes=False)
            else:
                url1 = "/".join(["", self.project_name.lower(), endpoint.Model.__name__.lower(), ""])
                url2 = url1 + "<id>/"
                self.api.add_resource(endpoint, url1, url2, endpoint=endpoint.__name__.lower(), strict_slashes=False)


def param_extrator(param_class):
    if not issubclass(param_class, Param):
        raise TypeError("param_class should be subclass of Param")

    def wrapper(func):
        @wraps(func)
        def wrapper2(*args, **kwargs):
            if request.content_type == "application/json":
                params = request.get_json()
            elif request.content_type == "application/json":
                params = request.get_data()
            param = param_class(**params)
            args = list(args)
            args.insert(1, param)
            return func(*args, **kwargs)

        return wrapper2

    return wrapper


import os
import subprocess


class Result(object):

    def __init__(self, output, err=None, status_code=0):
        self.stdout = output
        self.err = err
        self.status_code = status_code


class Cmd(object):

    def __init__(self, cmd):
        self.cmd = [cmd]

    def __call__(self, *args):
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            self.cmd_line = " ".join(self.cmd + list(args[0]))
        else:
            self.cmd_line = " ".join(self.cmd + list(args))
        process = subprocess.Popen(self.cmd_line, shell=True, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = process.communicate()
        status_code = process.wait()
        if not status_code:
            print("Command Success: {}".format(self.cmd_line))
        else:
            print("Command Failed: {}, ERROR_INFO: {}".format(self.cmd_line, err))
        return Result(out, err, status_code)

sys_apidoc = Cmd("apidoc")



from flask_autoapi.storage.handler import (FileHandler, MinioHandler,
                                           QiniuHandler, TencentHandler)


class Client(object):
    def __init__(self):
        self.conn = None
        self.kind = None
        self.handler_class = None

    def __getattribute__(self, name):
        if hasattr(self, name):
            return getattr(self, name)
        return getattr(self.conn, name)


class FileClient(Client):
    pass


class MinioClient(Client):
    def __init__(self, url, access_key, secret_key, secure):
        from minio import Minio
        self.kind = "minio"
        self.conn = Minio(
            url,
            access_key=access_key,
            secret_key=secret_key,
            secure=secure,
        )
        self.handler_class = MinioHandler


class QiniuClient(Client):
    """七牛对象存储"""
    def __init__(self, url, access_key, secret_key):
        from qiniu import Auth, put_data
        self.kind = "qiniu"
        self.conn = Auth(access_key, secret_key)
        self.qiniu_url = url
        self.put_data = put_data
        self.handler_class = QiniuHandler


class TencentClient(Client):
    """腾讯云对象存储"""
    def __init__(self, secret_id, secret_key, region):
        from qcloud_cos import CosConfig
        from qcloud_cos import CosS3Client
        self.kind = "tencent"
        config = CosConfig(
            Region=region,
            SecretId=secret_id,
            SecretKey=secret_key,
        )
        self.conn = CosS3Client(config)
        self.handler_class = TencentHandler



import os
import json
import peewee
import inspect
from jinja2 import Template
from flask_script import Command

from flask_autoapi.utils.filter import standard_type, str_align, get_example, is_mtom, mtom_fields, method_field_example, is_foreign
from flask_autoapi.utils.cmd import sys_apidoc
from flask_autoapi.endpoint import BaseEndpoint, BaseListEndpoint

# doc_tmpl = """
#         @api {{Method}} {{Url}} {{Title}}
#         @apiGroup {{Group}}

#         @apiExample 参数
#         {%- for field in Fields %}
#         {{ str_align(standard_type(field.field_type))}} \t {{ str_align(field.name, 15) }}  # {% if field.null is sameas true %} 非必填项 {% else %} 必填项 {% endif %} {% if field.verbose_name %}, {{field.verbose_name}} {% endif %}{% endfor %}

#         @apiExample 返回值
# {{DATA}}
# """


class GenerateDoc(Command):
    def __init__(self, api, diy_endpoint_list=None, static_folder="static"):
        self.api = api
        self.static_folder = static_folder
        self.doc_model_list = api.model_list
        self.diy_endpoint_list = diy_endpoint_list if diy_endpoint_list else []
        self.docs_folder = os.path.join(self.static_folder, "docs")

    def run(self):
        if not os.path.exists(self.static_folder):
            os.makedirs(self.static_folder)
        docs = [
            {
                "method": "get",
                "content": BaseEndpoint.get.__doc__,
            },
            {
                "method": "post",
                "content": BaseEndpoint.post.__doc__,
            },
            {
                "method": "put",
                "content": BaseEndpoint.put.__doc__,
            },
            {
                "method": "delete",
                "content": BaseEndpoint.delete.__doc__,
            },
            {
                "method": "list",
                "content": BaseListEndpoint.get.__doc__,
            },
        ]

        f = open(os.path.join(self.static_folder, "doc.py"), "w+")
        # template = Template(doc_tmpl)
        # for endpoint in self.diy_endpoint_list:
        #     titles = getattr(endpoint, "api_title")
        #     if not titles:
        #         continue
        #     for method in ["get", "post", "put", "delete"]:
        #         if not hasattr(endpoint, method):
        #             continue
        #         handler = getattr(endpoint, method)
        #         param_class = inspect.getclosurevars(handler).nonlocals.get(
        #             "param_class")
        #         if not param_class:
        #             continue
        #         content = template.render(
        #             Url=self.api._custom_resources.get(endpoint.__name__),
        #             Method="{"+method.upper()+"}",
        #             Title=titles.get(method),
        #             Group=getattr(endpoint, "api_group", "Default"),
        #             Fields=param_class.fields.values(),
        #             str_align=str_align,
        #             standard_type=standard_type,
        #         )
        #         f.write('"""' + content + '\n"""\n')

        for endpoint in self.diy_endpoint_list:
            for method in ["get", "post", "put", "delete"]:
                if not hasattr(endpoint, method):
                    continue
                content = getattr(getattr(endpoint, method), "__doc__")
                if not content:
                    continue
                f.write('"""' + content + '\n"""\n')

        for model in self.doc_model_list:
            fields = model.get_display_fields()
            all_fields = model.get_fields() + list(
                model._meta.manytomany.values()) + list(
                    model._meta.method_fields.values())
            r = get_model_example(model)
            data = {"code": 0, "message": "", "data": r}
            list_data = {"code": 0, "message": "", "data": [r]}
            model._meta.api_methods = [
                method.upper() for method in model._meta.api_methods
            ]
            for doc in docs:
                if not (doc.get("content")
                        and doc["method"].upper() in model._meta.api_methods):
                    continue
                template = Template(doc.get("content"))
                content = template.render(
                    Fields=model.get_post_fields()
                    if doc["method"] in ("post", "put") else fields,
                    # AllFields=all_fields,
                    ModelName=model.__name__,
                    Title=model._meta.verbose_name,
                    Group=model._meta.group or model.__name__,
                    str_align=str_align,
                    standard_type=standard_type,
                    # get_example=get_example,
                    project_name=self.api.project_name,
                    # is_mtom=is_mtom,
                    # mtom_fields=mtom_fields,
                    # method_field_example=method_field_example,
                    DATA=json.dumps(data, indent=4, sort_keys=True),
                    LIST_DATA=json.dumps(list_data, indent=4, sort_keys=True),
                )
                f.write('"""' + content + '\n"""\n')
        f.close()
        sys_apidoc("-i", self.static_folder, "-o", self.docs_folder)


def get_model_example(model):
    data = {}
    all_fields = model.get_fields() + list(
        model._meta.manytomany.values()) + list(
            model._meta.method_fields.values())
    for field in all_fields:
        if is_foreign(field):
            data[field.name] = get_model_example(field.rel_model)
        elif is_mtom(field) and not field._is_backref:
            data[field.name] = [get_model_example(field.rel_model)]
        elif is_mtom(field):
            data[field.name] = get_example(standard_type(field.field_type))
        elif field.field_type == "METHOD":
            data[field.name] = method_field_example(field)
        else:
            data[field.name] = get_example(standard_type(field.field_type),
                                           field.choices)
    return data


# jinja2 filters
import uuid
import json
import peewee
from datetime import datetime

from flask_autoapi.utils.diyutils import datetime_to_str


def standard_type(t):
    if not (t and isinstance(t, str)):
        return t
    t = t.upper()
    if t == "CHAR(32)":
        return "uuid"
    if t == "VARCHAR" or t.find("CHAR")>=0:
        return "string"
    return t.lower()

def str_align(word, length=10):
    return word.ljust(length," ")

def get_example(tp, choices=None):
    if choices:
        return choices[0][0]
    if tp == "uuid":
        return str(uuid.uuid4())
    if tp == "string":
        return "123456"
    if tp == "datetime":
        t = datetime.now()
        return datetime_to_str(t)
    if tp in ("int", "float", "double", "decimal"):
        return 12
    if tp == "bool":
        return 1
    return tp

def is_mtom(field):
    return isinstance(field, peewee.ManyToManyField)

def is_foreign(field):
    return isinstance(field, peewee.ForeignKeyField)

def mtom_fields(field):
    model = field.rel_model
    fields = model.get_fields()
    # fields = model.get_fields() + list(model._meta.manytomany.values())
    return fields

def foreign_fields(field):
    model = field.rel_model
    fields = model.get_fields()
    return fields

def method_field_example(field):
    if not field.field_type == "METHOD":
        return None
    r = field.method_class.get_example()
    # if not isinstance(r, (list, tuple, dict)):
    #     return r
    # r = json.dumps(r, sort_keys=True, indent=4)
    # result = ""
    # for line in r.split("\n"):
    #     result += "\n                " + line
    # return result
    return r